import time
from collections import OrderedDict
from collections.abc import Callable

import torch
from flwr.app import ArrayRecord, ConfigRecord, Context, MetricRecord
from flwr.serverapp import Grid, ServerApp
from flwr.serverapp.strategy import FedAvg
from torch import nn

from .models import get_model
from .task import load_centralized_dataset, test

# Create ServerApp
app = ServerApp()

device = "cpu"
if torch.cuda.is_available():
    device = "cuda"
elif torch.backends.mps.is_available():
    device = "mps"


@app.main()
def main(grid: Grid, context: Context) -> None:
    """Main entry point for the ServerApp."""

    # Read run config
    fraction_evaluate: float = float(context.run_config["fraction-evaluate"])
    num_rounds: int = int(context.run_config["num-server-rounds"])
    batch_size: int = int(context.run_config["batch-size"])
    lr: float = float(context.run_config["learning-rate"])

    # Load global model
    global_model = get_model(str(context.run_config["model-name"]), num_classes=10)
    global_model_state_dict = global_model.state_dict()
    assert isinstance(global_model_state_dict, OrderedDict)
    arrays = ArrayRecord(global_model_state_dict)

    # Initialize FedAvg strategy
    strategy = FedAvg(fraction_evaluate=fraction_evaluate, min_evaluate_nodes=0)

    # Start strategy, run FedAvg for `num_rounds`
    start_time = time.perf_counter()
    _ = strategy.start(
        grid=grid,
        initial_arrays=arrays,
        train_config=ConfigRecord({"lr": lr}),
        num_rounds=num_rounds,
        evaluate_fn=get_evaluate_fn(global_model, batch_size),
    )
    end_time = time.perf_counter()
    print(f"BENCHMARK_RESULT_TIME: {end_time - start_time:.4f}")


def get_evaluate_fn(
    model: nn.Module, batch_size: int
) -> Callable[[int, ArrayRecord], MetricRecord]:
    def evaluate(server_round: int, arrays: ArrayRecord) -> MetricRecord:
        """Evaluate model on central data."""

        # Load the model and initialize it with the received weights
        model.load_state_dict(arrays.to_torch_state_dict())
        model.to(device)

        # Load entire test set
        test_dataloader = load_centralized_dataset(batch_size)

        # Evaluate the global model on the test set
        test_loss, test_acc = test(model, test_dataloader, device)

        # Return the evaluation metrics
        return MetricRecord({"accuracy": test_acc, "loss": test_loss})

    return evaluate



# ruff: noqa: E402
import logging
import warnings

from pydantic.warnings import UnsupportedFieldAttributeWarning

warnings.filterwarnings("ignore", category=UnsupportedFieldAttributeWarning)

import sys
from datetime import datetime
from pathlib import Path
from typing import Annotated

import torch
import typer
import wandb
from blazefl.reproducibility import setup_reproducibility

from algorithm import DSFLBaseServerHandler, DSFLThreadPoolClientTrainer
from dataset import DSFLPartitionedDataset
from models import DSFLModelName, DSFLModelSelector


class DSFLPipeline:
    def __init__(
        self,
        handler: DSFLBaseServerHandler,
        trainer: DSFLThreadPoolClientTrainer,
        run: wandb.Run,
    ) -> None:
        self.handler = handler
        self.trainer = trainer
        self.run = run

    def main(self):
        while not self.handler.if_stop():
            round_ = self.handler.round
            # server side
            sampled_clients = self.handler.sample_clients()
            broadcast = self.handler.downlink_package()

            # client side
            self.trainer.local_process(broadcast, sampled_clients)
            uploads = self.trainer.uplink_package()

            # server side
            for pack in uploads:
                self.handler.load(pack)

            summary = self.handler.get_summary()
            self.run.log(summary, step=round_)
            formatted_summary = ", ".join(f"{k}: {v:.3f}" for k, v in summary.items())
            logging.info(f"round: {round_}, {formatted_summary}")

        logging.info("done!")


def setup_logging() -> None:
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.INFO)
    formatter = logging.Formatter(
        "%(asctime)s [%(levelname)s] %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)


def main(
    model_name: Annotated[
        DSFLModelName, typer.Option(help="Name of the model to be used.")
    ] = DSFLModelName.CNN,
    num_clients: Annotated[
        int, typer.Option(help="Total number of clients in the federation.")
    ] = 100,
    global_round: Annotated[
        int, typer.Option(help="Total number of federated learning rounds.")
    ] = 5,
    sample_ratio: Annotated[
        float, typer.Option(help="Fraction of clients to sample in each round.")
    ] = 1.0,
    partition: Annotated[
        str,
        typer.Option(help="Dataset partition strategy ('client_inner_dirichlet')."),
    ] = "client_inner_dirichlet",
    dir_alpha: Annotated[
        float,
        typer.Option(help="Alpha for Dirichlet distribution based partitioning."),
    ] = 1.0,
    seed: Annotated[int, typer.Option(help="Seed for reproducibility.")] = 42,
    epochs: Annotated[
        int, typer.Option(help="Number of local training epochs per client.")
    ] = 5,
    lr: Annotated[
        float, typer.Option(help="Learning rate for the client optimizer.")
    ] = 0.1,
    batch_size: Annotated[
        int, typer.Option(help="Batch size for local training.")
    ] = 50,
    num_parallels: Annotated[
        int,
        typer.Option(help="Number of parallel processes for training."),
    ] = 10,
    dataset_root_dir: Annotated[
        Path, typer.Option(help="Root directory for the dataset.")
    ] = Path("/tmp/step-by-step-dsfl/dataset"),
    state_dir_base: Annotated[
        Path, typer.Option(help="Directory path for saving data between processes.")
    ] = Path("/tmp/step-by-step-dsfl/state"),
    open_size: Annotated[int, typer.Option(help="Size of the open dataset.")] = 10000,
    kd_epochs: Annotated[
        int, typer.Option(help="Number of knowledge distillation epochs.")
    ] = 5,
    kd_batch_size: Annotated[
        int, typer.Option(help="Batch size for knowledge distillation.")
    ] = 50,
    kd_lr: Annotated[
        float, typer.Option(help="Learning rate for knowledge distillation.")
    ] = 0.1,
    era_temperature: Annotated[
        float, typer.Option(help="Temperature for Entropy Reduction Aggregation.")
    ] = 0.1,
    open_size_per_round: Annotated[
        int, typer.Option(help="Size of the open dataset to use per round.")
    ] = 1000,
) -> None:
    device = "cpu"
    if torch.cuda.is_available():
        device = "cuda"
    elif torch.backends.mps.is_available():
        device = "mps"

    config = locals()
    run = wandb.init(mode="offline", config=config)

    setup_logging()
    logging.info("\n" + "\n".join([f"  {k}: {v}" for k, v in config.items()]))

    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    dataset_split_dir = dataset_root_dir / timestamp
    state_dir = state_dir_base / timestamp

    setup_reproducibility(seed)

    dataset = DSFLPartitionedDataset(
        root=dataset_root_dir,
        path=dataset_split_dir,
        num_clients=num_clients,
        dir_alpha=dir_alpha,
        seed=seed,
        partition=partition,
        open_size=open_size,
    )
    model_selector = DSFLModelSelector(num_classes=10, seed=seed)

    handler = DSFLBaseServerHandler(
        model_selector=model_selector,
        model_name=model_name,
        dataset=dataset,
        global_round=global_round,
        num_clients=num_clients,
        kd_epochs=kd_epochs,
        kd_batch_size=kd_batch_size,
        kd_lr=kd_lr,
        era_temperature=era_temperature,
        open_size_per_round=open_size_per_round,
        device=device,
        sample_ratio=sample_ratio,
        seed=seed,
    )
    trainer = DSFLThreadPoolClientTrainer(
        model_selector=model_selector,
        model_name=model_name,
        dataset=dataset,
        state_dir=state_dir,
        seed=seed,
        device=device,
        num_clients=num_clients,
        epochs=epochs,
        batch_size=batch_size,
        lr=lr,
        kd_epochs=kd_epochs,
        kd_batch_size=kd_batch_size,
        kd_lr=kd_lr,
        num_parallels=num_parallels,
    )
    pipeline = DSFLPipeline(handler=handler, trainer=trainer, run=run)

    try:
        pipeline.main()
    except KeyboardInterrupt:
        logging.info("KeyboardInterrupt")


if __name__ == "__main__":
    typer.run(main)



from collections import OrderedDict

import torch
from flwr.app import ArrayRecord, Context, Message, MetricRecord, RecordDict
from flwr.clientapp import ClientApp

from example.task import train as train_fn

from .models import get_model
from .task import load_data

# Flower ClientApp
app = ClientApp()

device = "cpu"
if torch.cuda.is_available():
    device = "cuda"
elif torch.backends.mps.is_available():
    device = "mps"


@app.train()
def train(msg: Message, context: Context):
    """Train the model on local data."""

    # Load the model and initialize it with the received weights
    model = get_model(str(context.run_config["model-name"]), num_classes=10)
    arrays = msg.content["arrays"]
    assert isinstance(arrays, ArrayRecord)
    model.load_state_dict(arrays.to_torch_state_dict())
    model.to(device)

    # Load the data
    partition_id = int(context.node_config["partition-id"])
    num_partitions = int(context.node_config["num-partitions"])
    batch_size = int(context.run_config["batch-size"])
    trainloader = load_data(partition_id, num_partitions, batch_size)

    # Call the training function
    train_loss = train_fn(
        model,
        trainloader,
        context.run_config["local-epochs"],
        msg.content["config"]["lr"],
        device,
    )

    # Construct and return reply Message
    model_state_dict = model.state_dict()
    assert isinstance(model_state_dict, OrderedDict)
    model_record = ArrayRecord(model_state_dict)
    metrics = {
        "train_loss": train_loss,
        "num-examples": len(trainloader.dataset),  # pyright: ignore[reportArgumentType]
    }
    metric_record = MetricRecord(metrics)
    content = RecordDict({"arrays": model_record, "metrics": metric_record})
    return Message(content=content, reply_to=msg)



import os
import random
from dataclasses import dataclass

import numpy as np
import torch


def setup_reproducibility(seed: int) -> None:
    """
    Configures the environment-level settings for deterministic behavior.

    This function sets the `PYTHONHASHSEED` for consistent hash-based operations
    and configures PyTorch's cuDNN backend to use deterministic algorithms.
    Call this at the start of your script for a stable environment.

    Args:
        seed: The seed value to use for the hash seed.
    """
    os.environ["PYTHONHASHSEED"] = str(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False


@dataclass
class RNGSuite:
    """
    A container for a suite of isolated random number generators.

    This class holds independent, seeded generator objects for each library.
    Using an `RNGSuite` instance allows for randomness control that is self-contained
    and does not interfere with the global state, making it suitable for components or
    libraries that need their own random stream.

    Attributes:
        python: An isolated `random.Random` generator.
        numpy: An isolated `numpy.random.Generator` instance.
        torch_cpu: A `torch.Generator` for CPU operations.
        torch_cuda: A `torch.Generator` for CUDA operations, if available.
    """

    python: random.Random
    numpy: np.random.Generator
    torch_cpu: torch.Generator
    torch_cuda: torch.Generator | None = None


def create_rng_suite(seed: int) -> RNGSuite:
    """
    Creates a new suite of isolated random number generators from a single seed.

    This is a convenience factory function to instantiate `RNGSuite` with all its
    generators properly seeded and ready for use.

    Args:
        seed: The master seed to initialize all generators in the suite.

    Returns:
        A new `RNGSuite` instance.
    """
    python_rng = random.Random(seed)
    numpy_rng = np.random.default_rng(seed)
    torch_cpu_rng = torch.Generator(device="cpu").manual_seed(seed)

    torch_cuda_rng = None
    if torch.cuda.is_available():
        torch_cuda_rng = torch.Generator("cuda").manual_seed(seed)

    return RNGSuite(
        python=python_rng,
        numpy=numpy_rng,
        torch_cpu=torch_cpu_rng,
        torch_cuda=torch_cuda_rng,
    )



from enum import StrEnum

import torch

from src.blazefl.core import ModelSelector


class DummyModelName(StrEnum):
    LINEAR = "linear"
    CONV = "conv"


class DummyModelSelector(ModelSelector[DummyModelName]):
    def select_model(self, model_name: DummyModelName) -> torch.nn.Module:
        match model_name:
            case DummyModelName.LINEAR:
                return torch.nn.Linear(10, 5)
            case DummyModelName.CONV:
                return torch.nn.Conv2d(1, 3, 3)


def test_model_selector_subclass() -> None:
    selector = DummyModelSelector()
    model = selector.select_model(DummyModelName.LINEAR)
    assert isinstance(model, torch.nn.Linear)
    model = selector.select_model(DummyModelName.CONV)
    assert isinstance(model, torch.nn.Conv2d)



import signal
import threading
from collections.abc import Iterable
from concurrent.futures import Future, ThreadPoolExecutor
from multiprocessing.pool import ApplyResult
from typing import Protocol, TypeVar

from blazefl.core.utils import process_tensors_in_object, reconstruct_from_shared_memory

UplinkPackage = TypeVar("UplinkPackage")
DownlinkPackage = TypeVar("DownlinkPackage", contravariant=True)


class BaseClientTrainer(Protocol[UplinkPackage, DownlinkPackage]):
    """
    Abstract base class for serial client training in federated learning.

    This class defines the interface for training clients in a serial manner,
    where each client is processed one after the other.

    Raises:
        NotImplementedError: If the methods are not implemented in a subclass.
    """

    def uplink_package(self) -> list[UplinkPackage]:
        """
        Prepare the data package to be sent from the client to the server.

        Returns:
            list[UplinkPackage]: A list of data packages prepared for uplink
            transmission.
        """
        ...

    def local_process(self, payload: DownlinkPackage, cid_list: list[int]) -> None:
        """
        Process the downlink payload from the server for a list of client IDs.

        Args:
            payload (DownlinkPackage): The data package received from the server.
            cid_list (list[int]): A list of client IDs to process.

        Returns:
            None
        """
        ...


ClientConfig = TypeVar("ClientConfig")


class ProcessPoolClientTrainer(
    BaseClientTrainer[UplinkPackage, DownlinkPackage],
    Protocol[UplinkPackage, DownlinkPackage, ClientConfig],
):
    """
    Abstract base class for parallel client training in federated learning.

    This class extends SerialClientTrainer to enable parallel processing of clients,
    allowing multiple clients to be trained concurrently.

    Attributes:
        num_parallels (int): Number of parallel processes to use for client training.
        device (str): The primary device to use for computation (e.g., "cpu", "cuda").
        device_count (int): The number of available CUDA devices, if `device` is "cuda".
        cache (list[UplinkPackage]): Cache to store uplink packages from clients.

    Raises:
        NotImplementedError: If the abstract methods are not implemented in a subclass.
    """

    num_parallels: int
    device: str
    device_count: int
    cache: list[UplinkPackage]
    stop_event: threading.Event

    def progress_fn(
        self,
        it: list[ApplyResult],
    ) -> Iterable[ApplyResult]:
        """
        A no-op progress function that can be overridden to provide custom
        progress tracking.

        Args:
            it (list[ApplyResult]): A list of ApplyResult objects.

        Returns:
            Iterable[ApplyResult]: The original iterable.
        """
        return it

    def get_client_config(self, cid: int) -> ClientConfig:
        """
        Retrieve the configuration for a given client ID.

        Args:
            cid (int): Client ID.

        Returns:
            ClientConfig: The configuration for the specified client.
        """
        ...

    def get_client_device(self, cid: int) -> str:
        """
        Retrieve the device to use for processing a given client.

        Args:
            cid (int): Client ID.

        Returns:
            str: The device to use for processing the client.
        """
        if self.device == "cuda":
            return f"cuda:{cid % self.device_count}"
        return self.device

    @staticmethod
    def worker(
        config: ClientConfig,
        payload: DownlinkPackage,
        device: str,
        stop_event: threading.Event,
        *,
        shm_buffer: UplinkPackage | None = None,
    ) -> UplinkPackage:
        """
        Process a single client's training task.

        This method is executed by each worker process in the pool.
        It handles loading client configuration and payload, performing
        the client-specific operations, and returning the result.

        Args:
            config (ClientConfig):
                The client's configuration data.
            payload (DownlinkPackage):
                The downlink payload from the server
            device (str): Device to use for processing (e.g., "cpu", "cuda:0").
            stop_event (threading.Event): Event to signal stopping the worker.
            shm_buffer (UplinkPackage | None):
                Optional shared memory buffer for the uplink package.

        Returns:
            UplinkPackage:
                The uplink package containing the client's results.
        """
        ...

    def prepare_uplink_package_buffer(self) -> UplinkPackage:
        raise NotImplementedError

    def local_process(self, payload: DownlinkPackage, cid_list: list[int]) -> None:
        """
        Manage the parallel processing of clients.

        This method distributes the processing of multiple clients across
        parallel processes, handling data saving, loading, and caching.

        Args:
            payload (DownlinkPackage): The data package received from the server.
            cid_list (list[int]): A list of client IDs to process.

        Returns:
            None
        """
        import torch.multiprocessing as mp

        shm_buffers = {}
        process_tensors_in_object(payload, mode="move")
        for cid in cid_list:
            buffer = self.prepare_uplink_package_buffer()
            process_tensors_in_object(buffer, mode="move")
            shm_buffers[cid] = buffer

        self.stop_event.clear()
        pool = mp.Pool(
            processes=self.num_parallels,
            initializer=signal.signal,
            initargs=(signal.SIGINT, signal.SIG_IGN),
        )
        try:
            jobs: list[ApplyResult] = []
            for cid in cid_list:
                config = self.get_client_config(cid)
                device = self.get_client_device(cid)
                jobs.append(
                    pool.apply_async(
                        self.worker,
                        (
                            config,
                            payload,
                            device,
                            self.stop_event,
                        ),
                        kwds={
                            "shm_buffer": shm_buffers.get(cid),
                        },
                    ),
                )

            for i, job in enumerate(self.progress_fn(jobs)):
                result = job.get()
                cid = cid_list[i]
                package = reconstruct_from_shared_memory(result, shm_buffers[cid])
                self.cache.append(package)
        finally:
            self.stop_event.set()
            pool.close()
            pool.join()


class ThreadPoolClientTrainer(
    BaseClientTrainer[UplinkPackage, DownlinkPackage],
    Protocol[UplinkPackage, DownlinkPackage],
):
    num_parallels: int
    device: str
    device_count: int
    cache: list[UplinkPackage]
    stop_event: threading.Event

    def progress_fn(
        self, it: list[Future[UplinkPackage]]
    ) -> Iterable[Future[UplinkPackage]]:
        """
        A no-op progress function that can be overridden to provide custom
        progress tracking.

        Args:
            it (list[Future[UplinkPackage]]): A list of Future objects
                representing the results of client processing.

        Returns:
            Iterable[Future[UplinkPackage]]: The original iterable.
        """
        return it

    def worker(
        self,
        cid: int,
        device: str,
        payload: DownlinkPackage,
        stop_event: threading.Event,
    ) -> UplinkPackage:
        """
        Process a single client's training task in a thread.

        Args:
            cid (int): The client ID.
            device (str): The device to use for processing this client.
            payload (DownlinkPackage): The data package received from the server.
            stop_event (threading.Event): Event to signal stopping the worker.

        Returns:
            UplinkPackage: The uplink package containing the client's results.
        """
        ...

    def get_client_device(self, cid: int) -> str:
        """
        Retrieve the device to use for processing a given client.

        Args:
            cid (int): Client ID.

        Returns:
            str: The device to use for processing the client.
        """
        if self.device == "cuda":
            return f"cuda:{cid % self.device_count}"
        return self.device

    def local_process(self, payload: DownlinkPackage, cid_list: list[int]) -> None:
        """
        Manage the parallel processing of clients using threads.

        This method distributes the processing of multiple clients across
        a pool of threads.

        Args:
            payload (DownlinkPackage): The data package received from the server.
            cid_list (list[int]): A list of client IDs to process.
        """
        self.stop_event.clear()
        executor = ThreadPoolExecutor(max_workers=self.num_parallels)
        try:
            futures: list[Future[UplinkPackage]] = []
            for cid in cid_list:
                device = self.get_client_device(cid)
                future = executor.submit(
                    self.worker,
                    cid,
                    device,
                    payload,
                    self.stop_event,
                )
                futures.append(future)

            for future in self.progress_fn(futures):
                result = future.result()
                self.cache.append(result)
        finally:
            self.stop_event.set()
            executor.shutdown(wait=True, cancel_futures=True)



import numpy as np
import numpy.typing as npt
import torch
from blazefl.reproducibility.generator import setup_reproducibility as setup_reproducibility
from dataclasses import dataclass
from typing import Any

def seed_everything(seed: int) -> None: ...

@dataclass
class RandomStateSnapshot:
    environ: str
    python: tuple[Any, ...]
    numpy: tuple[str, npt.NDArray[np.uint32], int, int, float]
    torch_cpu: torch.Tensor
    torch_cuda: torch.Tensor | None
    @classmethod
    def capture(cls) -> RandomStateSnapshot: ...
    @staticmethod
    def restore(snapshot: RandomStateSnapshot) -> None: ...



import torch
from enum import StrEnum
from typing import Protocol, TypeVar

ModelName = TypeVar('ModelName', bound=StrEnum, contravariant=True)

class ModelSelector(Protocol[ModelName]):
    def select_model(self, model_name: ModelName) -> torch.nn.Module: ...



from enum import StrEnum

import torch
from blazefl.core import ModelSelector
from torch import nn
from torchvision.models import resnet18

from models.cnn import CNN


class DSFLModelName(StrEnum):
    CNN = "CNN"
    RESNET18 = "RESNET18"


class DSFLModelSelector(ModelSelector[DSFLModelName]):
    def __init__(self, num_classes: int, seed: int) -> None:
        self.num_classes = num_classes
        self.seed = seed

    def select_model(self, model_name: DSFLModelName) -> nn.Module:
        with torch.random.fork_rng([]):
            torch.manual_seed(self.seed)
            match model_name:
                case DSFLModelName.CNN:
                    return CNN(num_classes=self.num_classes)
                case DSFLModelName.RESNET18:
                    return resnet18(num_classes=self.num_classes)



from blazefl.reproducibility.generator import RNGSuite as RNGSuite, create_rng_suite as create_rng_suite, setup_reproducibility as setup_reproducibility
from blazefl.reproducibility.snapshot import RandomStateSnapshot as RandomStateSnapshot, seed_everything as seed_everything

__all__ = ['seed_everything', 'setup_reproducibility', 'RandomStateSnapshot', 'create_rng_suite', 'RNGSuite']



import random

import numpy as np
import pytest
import torch

from src.blazefl.reproducibility import RandomStateSnapshot, seed_everything


def test_seed_everything() -> None:
    seed = 42
    seed_everything(seed)

    py_rand_val = random.random()
    np_rand_val = np.random.rand()
    torch_rand_val = torch.rand(1)

    seed_everything(seed)
    assert py_rand_val == random.random()
    assert np.allclose(np_rand_val, np.random.rand())
    assert torch.allclose(torch_rand_val, torch.rand(1))


def test_random_state_snapshot_cpu() -> None:
    device = "cpu"
    seed = 123
    seed_everything(seed)

    _ = random.random()
    _ = np.random.rand()
    _ = torch.rand(1, device=device)

    snapshot = RandomStateSnapshot.capture()

    py_rand_val_before = random.random()
    np_rand_val_before = np.random.rand()
    torch_rand_val_before = torch.rand(1, device=device)

    RandomStateSnapshot.restore(snapshot)

    assert py_rand_val_before == random.random()
    assert np.allclose(np_rand_val_before, np.random.rand())
    assert torch.allclose(torch_rand_val_before, torch.rand(1, device=device))


@pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA not available")
def test_random_state_snapshot_cuda() -> None:
    device = "cuda"
    seed = 123
    seed_everything(seed)

    _ = random.random()
    _ = np.random.rand()
    _ = torch.rand(1, device=device)

    snapshot = RandomStateSnapshot.capture()

    py_rand_val_before = random.random()
    np_rand_val_before = np.random.rand()
    torch_rand_val_before = torch.rand(1, device=device)

    RandomStateSnapshot.restore(snapshot)

    assert py_rand_val_before == random.random()
    assert np.allclose(np_rand_val_before, np.random.rand())
    assert torch.allclose(torch_rand_val_before, torch.rand(1, device=device))



import threading
from dataclasses import dataclass

import pytest
import torch
import torch.multiprocessing as mp

from blazefl.core.utils import SHMHandle
from src.blazefl.core import ProcessPoolClientTrainer


@dataclass
class UplinkPackage:
    cid: int
    message: str
    tensor: torch.Tensor | SHMHandle


@dataclass
class DownlinkPackage:
    message: str


@dataclass
class ClientConfig:
    cid: int


class DummyProcessPoolClientTrainer(
    ProcessPoolClientTrainer[UplinkPackage, DownlinkPackage, ClientConfig]
):
    def __init__(
        self,
        num_parallels: int,
        device: str,
    ):
        self.num_parallels = num_parallels
        self.device = device
        if self.device == "cuda":
            self.device_count = torch.cuda.device_count()
        self.cache: list[UplinkPackage] = []
        self.manager = mp.Manager()
        self.stop_event = self.manager.Event()

    def uplink_package(self) -> list[UplinkPackage]:
        return self.cache

    def get_client_config(self, cid: int) -> ClientConfig:
        return ClientConfig(cid=cid)

    def prepare_uplink_package_buffer(self) -> UplinkPackage:
        return UplinkPackage(cid=-1, message="", tensor=torch.zeros(1))

    @staticmethod
    def worker(
        config: ClientConfig,
        payload: DownlinkPackage,
        device: str,
        stop_event: threading.Event,
        *,
        shm_buffer: UplinkPackage | None = None,
    ) -> UplinkPackage:
        _ = stop_event
        _ = device
        dummy_uplink_package = UplinkPackage(
            cid=config.cid,
            tensor=torch.rand(1),
            message=payload.message + "<client_to_server>",
        )

        assert shm_buffer is not None
        shm_buffer.tensor = dummy_uplink_package.tensor
        dummy_uplink_package.tensor = SHMHandle()
        return dummy_uplink_package


@pytest.mark.parametrize("num_parallels", [1, 2, 4])
@pytest.mark.parametrize("cid_list", [[], [42], [0, 1, 2]])
def test_process_pool_client_trainer(num_parallels: int, cid_list: list[int]) -> None:
    trainer = DummyProcessPoolClientTrainer(
        num_parallels=num_parallels,
        device="cpu",
    )

    dummy_payload = DownlinkPackage(message="<server_to_client>")

    trainer.local_process(dummy_payload, cid_list)

    assert len(trainer.cache) == len(cid_list)
    for i, cid in enumerate(cid_list):
        result = trainer.cache[i]
        assert result.cid == cid
        assert result.message == "<server_to_client><client_to_server>"

    package = trainer.uplink_package()
    assert len(package) == len(cid_list)

    for i, cid in enumerate(cid_list):
        result = package[i]
        assert result.cid == cid
        assert result.message == "<server_to_client><client_to_server>"



import os
import signal
import time
from contextlib import suppress
from enum import StrEnum

import psutil
import pytest
import torch
import torch.multiprocessing as mp
from torch.utils.data import DataLoader, Dataset

from src.blazefl.contrib.fedavg import (
    FedAvgBaseClientTrainer,
    FedAvgBaseServerHandler,
    FedAvgDownlinkPackage,
    FedAvgProcessPoolClientTrainer,
    FedAvgThreadPoolClientTrainer,
)
from src.blazefl.core import ModelSelector, PartitionedDataset


class DummyModelName(StrEnum):
    DUMMY = "dummy"


class DummyModelSelector(ModelSelector[DummyModelName]):
    def select_model(self, model_name: DummyModelName) -> torch.nn.Module:
        match model_name:
            case DummyModelName.DUMMY:
                return torch.nn.Sequential(torch.nn.Flatten(), torch.nn.Linear(4, 2))


class DummyDataset(Dataset):
    def __init__(self, size=10):
        self.data = torch.randn(size, 2, 2)
        self.targets = torch.randint(0, 2, (size,))

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):
        return self.data[index], self.targets[index]


class DummyDataType(StrEnum):
    TRAIN = "train"
    TEST = "test"


class DummyPartitionedDataset(PartitionedDataset[DummyDataType]):
    def __init__(self, num_clients: int, size_per_client: int):
        self.num_clients = num_clients
        self.datasets = []
        for cid in range(num_clients):
            g = torch.Generator().manual_seed(cid)
            data = torch.randn((size_per_client, 2, 2), generator=g)
            targets = torch.randint(0, 2, (size_per_client,), generator=g)
            dataset = torch.utils.data.TensorDataset(data, targets)
            self.datasets.append(dataset)

    def get_dataset(self, type_: DummyDataType, cid: int | None):
        match type_:
            case DummyDataType.TRAIN | DummyDataType.TEST:
                if cid is None:
                    cid = 0
                return self.datasets[cid]

    def get_dataloader(
        self,
        type_: DummyDataType,
        cid: int | None,
        batch_size: int | None,
        generator: torch.Generator | None = None,
    ):
        dataset = self.get_dataset(type_, cid)
        if batch_size is None:
            batch_size = len(dataset)
        return DataLoader(
            dataset,
            batch_size=batch_size,
            generator=generator,
        )


@pytest.fixture
def model_selector():
    return DummyModelSelector()


@pytest.fixture
def partitioned_dataset():
    return DummyPartitionedDataset(num_clients=10, size_per_client=10)


@pytest.fixture
def device():
    return "cuda" if torch.cuda.is_available() else "cpu"


@pytest.fixture
def tmp_state_dir(tmp_path):
    state_dir = tmp_path / "state"
    return state_dir


def test_base_server_and_base_trainer_integration(
    model_selector, partitioned_dataset, device
):
    model_name = DummyModelName.DUMMY
    global_round = 1
    num_clients = 3
    sample_ratio = 1.0
    epochs = 1
    batch_size = 2
    lr = 0.01
    seed = 42

    server = FedAvgBaseServerHandler(
        model_selector=model_selector,
        model_name=model_name,
        dataset=partitioned_dataset,
        global_round=global_round,
        num_clients=num_clients,
        sample_ratio=sample_ratio,
        device=device,
        batch_size=batch_size,
        seed=seed,
    )

    trainer = FedAvgBaseClientTrainer(
        model_selector=model_selector,
        model_name=model_name,
        dataset=partitioned_dataset,
        device=device,
        num_clients=num_clients,
        epochs=epochs,
        batch_size=batch_size,
        lr=lr,
        seed=seed,
    )

    cids = server.sample_clients()
    assert len(cids) == num_clients
    downlink = server.downlink_package()
    trainer.local_process(downlink, cids)
    uplinks = trainer.uplink_package()
    assert len(uplinks) == num_clients

    done = False
    for pkg in uplinks:
        done = server.load(pkg)
    assert done is True
    assert server.round == 1

    assert server.if_stop() is True


def _run_process_pool_trainer(
    trainer_init_args: dict, downlink: FedAvgDownlinkPackage, cids: list[int]
) -> None:
    trainer = FedAvgProcessPoolClientTrainer(**trainer_init_args)
    with suppress(KeyboardInterrupt):
        trainer.local_process(downlink, cids)


def test_base_handler_and_process_pool_trainer_integration(
    model_selector, partitioned_dataset, device, tmp_state_dir
):
    mp.set_start_method("spawn", force=True)

    model_name = DummyModelName.DUMMY
    global_round = 2
    num_clients = 10
    sample_ratio = 1.0
    epochs = 1
    batch_size = 2
    lr = 0.01
    seed = 42
    num_parallels = 2

    server = FedAvgBaseServerHandler(
        model_selector=model_selector,
        model_name=model_name,
        dataset=partitioned_dataset,
        global_round=global_round,
        num_clients=num_clients,
        sample_ratio=sample_ratio,
        device=device,
        batch_size=batch_size,
        seed=seed,
    )

    trainer = FedAvgProcessPoolClientTrainer(
        model_selector=model_selector,
        model_name=model_name,
        state_dir=tmp_state_dir,
        dataset=partitioned_dataset,
        device=device,
        num_clients=num_clients,
        epochs=epochs,
        batch_size=batch_size,
        lr=lr,
        seed=seed,
        num_parallels=num_parallels,
    )

    for round_ in range(1, global_round + 1):
        cids = server.sample_clients()
        downlink = server.downlink_package()
        trainer.local_process(downlink, cids)
        uplinks = trainer.uplink_package()
        assert len(uplinks) == num_clients

        done = False
        for pkg in uplinks:
            done = server.load(pkg)
        assert done is True
        assert server.round == round_

    assert server.if_stop() is True


def test_base_handler_and_process_pool_trainer_integration_keyboard_interrupt(
    model_selector, partitioned_dataset, device, tmp_state_dir
):
    mp.set_start_method("spawn", force=True)

    model_name = DummyModelName.DUMMY
    global_round = 1
    num_clients = 10
    sample_ratio = 1.0
    epochs = 10**5
    batch_size = 2
    lr = 0.01
    seed = 42
    num_parallels = 2

    server = FedAvgBaseServerHandler(
        model_selector=model_selector,
        model_name=model_name,
        dataset=partitioned_dataset,
        global_round=global_round,
        num_clients=num_clients,
        sample_ratio=sample_ratio,
        device=device,
        batch_size=batch_size,
        seed=seed,
    )

    trainer_init_args = {
        "model_selector": model_selector,
        "model_name": model_name,
        "state_dir": tmp_state_dir,
        "dataset": partitioned_dataset,
        "device": device,
        "num_clients": num_clients,
        "epochs": epochs,
        "batch_size": batch_size,
        "lr": lr,
        "seed": seed,
        "num_parallels": num_parallels,
    }

    cids = server.sample_clients()
    downlink = server.downlink_package()

    proc = mp.Process(
        target=_run_process_pool_trainer,
        args=(trainer_init_args, downlink, cids),
    )
    proc.start()
    assert proc.pid is not None

    time.sleep(5)
    assert proc.is_alive()

    os.kill(proc.pid, signal.SIGINT)

    proc.join(timeout=10)
    assert not proc.is_alive()


def test_base_handler_and_thread_pool_trainer_integration(
    model_selector, partitioned_dataset, device
):
    mp.set_start_method("spawn", force=True)

    model_name = DummyModelName.DUMMY
    global_round = 2
    num_clients = 10
    sample_ratio = 1.0
    epochs = 1
    batch_size = 2
    lr = 0.01
    seed = 42
    num_parallels = 2

    server = FedAvgBaseServerHandler(
        model_selector=model_selector,
        model_name=model_name,
        dataset=partitioned_dataset,
        global_round=global_round,
        num_clients=num_clients,
        sample_ratio=sample_ratio,
        device=device,
        batch_size=batch_size,
        seed=seed,
    )

    trainer = FedAvgThreadPoolClientTrainer(
        model_selector=model_selector,
        model_name=model_name,
        dataset=partitioned_dataset,
        device=device,
        num_clients=num_clients,
        epochs=epochs,
        batch_size=batch_size,
        lr=lr,
        seed=seed,
        num_parallels=num_parallels,
    )

    for round_ in range(1, global_round + 1):
        cids = server.sample_clients()
        downlink = server.downlink_package()
        trainer.local_process(downlink, cids)
        uplinks = trainer.uplink_package()
        assert len(uplinks) == num_clients

        done = False
        for pkg in uplinks:
            done = server.load(pkg)
        assert done is True
        assert server.round == round_

    assert server.if_stop() is True


def _run_thread_pool_trainer(
    trainer_init_args: dict, downlink: FedAvgDownlinkPackage, cids: list[int]
) -> None:
    trainer = FedAvgThreadPoolClientTrainer(**trainer_init_args)
    with suppress(KeyboardInterrupt):
        trainer.local_process(downlink, cids)


def test_base_handler_and_thread_pool_trainer_integration_keyboard_interrupt(
    model_selector,
    partitioned_dataset,
    device,
):
    mp.set_start_method("spawn", force=True)

    model_name = DummyModelName.DUMMY
    global_round = 1
    num_clients = 10
    sample_ratio = 1.0
    epochs = 10**5
    batch_size = 2
    lr = 0.01
    seed = 42
    num_parallels = 2

    server = FedAvgBaseServerHandler(
        model_selector=model_selector,
        model_name=model_name,
        dataset=partitioned_dataset,
        global_round=global_round,
        num_clients=num_clients,
        sample_ratio=sample_ratio,
        device=device,
        batch_size=batch_size,
        seed=seed,
    )

    trainer_args = {
        "model_selector": model_selector,
        "model_name": model_name,
        "dataset": partitioned_dataset,
        "device": device,
        "num_clients": num_clients,
        "epochs": epochs,
        "batch_size": batch_size,
        "lr": lr,
        "seed": seed,
        "num_parallels": num_parallels,
    }

    cids = server.sample_clients()
    downlink = server.downlink_package()

    proc = mp.Process(
        target=_run_thread_pool_trainer, args=(trainer_args, downlink, cids)
    )
    proc.start()
    assert proc.pid is not None

    p = psutil.Process(proc.pid)
    timeout = 5
    while p.num_threads() < num_parallels + 1:
        if not p.is_running() or time.time() - p.create_time() > timeout:
            pytest.fail(
                f"Process did not spawn {num_parallels} threads within {timeout}s"
            )
        time.sleep(0.1)
    assert proc.is_alive()
    time.sleep(3)

    os.kill(proc.pid, signal.SIGINT)

    proc.join(timeout=5)
    assert not proc.is_alive()



from enum import StrEnum

import torch
from blazefl.core import ModelSelector
from torch import nn
from torch.nn import functional as F
from torchvision.models import resnet18


class FedAvgModelName(StrEnum):
    CNN = "CNN"
    RESNET18 = "RESNET18"


class FedAvgModelSelector(ModelSelector[FedAvgModelName]):
    def __init__(self, num_classes: int, seed: int) -> None:
        self.num_classes = num_classes
        self.seed = seed

    def select_model(self, model_name: FedAvgModelName) -> nn.Module:
        with torch.random.fork_rng([]):
            torch.manual_seed(self.seed)
            match model_name:
                case FedAvgModelName.CNN:
                    return CNN(num_classes=self.num_classes)
                case FedAvgModelName.RESNET18:
                    return resnet18(num_classes=self.num_classes)


class CNN(nn.Module):
    """
    Based on
    https://pytorch.org/tutorials/beginner/blitz/cifar10_tutorial.html
    with slight modifications.
    """

    def __init__(self, num_classes: int) -> None:
        super().__init__()
        self.conv1 = nn.Conv2d(3, 6, 5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = nn.Conv2d(6, 16, 5)
        self.fc1 = nn.Linear(16 * 5 * 5, 120)
        self.fc2 = nn.Linear(120, 84)
        self.fc3 = nn.Linear(84, num_classes)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = torch.flatten(x, 1)  # flatten all dimensions except batch
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x



"""
Core module of BlazeFL framework.

This module imports and defines the core components of the BlazeFL framework,
including client trainers, model selectors, partitioned datasets, and server handlers.
"""

from blazefl.core.client_trainer import (
    BaseClientTrainer,
    ProcessPoolClientTrainer,
    ThreadPoolClientTrainer,
)
from blazefl.core.model_selector import ModelSelector
from blazefl.core.partitioned_dataset import FilteredDataset, PartitionedDataset
from blazefl.core.server_handler import BaseServerHandler
from blazefl.core.utils import (
    SHMHandle,
    deserialize_model,
    process_tensors_in_object,
    reconstruct_from_shared_memory,
    serialize_model,
)

__all__ = [
    "BaseClientTrainer",
    "FilteredDataset",
    "ProcessPoolClientTrainer",
    "ThreadPoolClientTrainer",
    "ModelSelector",
    "PartitionedDataset",
    "BaseServerHandler",
    "serialize_model",
    "deserialize_model",
    "process_tensors_in_object",
    "reconstruct_from_shared_memory",
    "SHMHandle",
]



"""
Reproducibility module for BlazeFL.
"""

from blazefl.reproducibility.generator import (
    RNGSuite,
    create_rng_suite,
    setup_reproducibility,
)
from blazefl.reproducibility.snapshot import (
    RandomStateSnapshot,
    seed_everything,
)

__all__ = [
    "seed_everything",
    "setup_reproducibility",
    "RandomStateSnapshot",
    "create_rng_suite",
    "RNGSuite",
]



from blazefl.core.client_trainer import BaseClientTrainer as BaseClientTrainer, ProcessPoolClientTrainer as ProcessPoolClientTrainer, ThreadPoolClientTrainer as ThreadPoolClientTrainer
from blazefl.core.model_selector import ModelSelector as ModelSelector
from blazefl.core.partitioned_dataset import FilteredDataset as FilteredDataset, PartitionedDataset as PartitionedDataset
from blazefl.core.server_handler import BaseServerHandler as BaseServerHandler
from blazefl.core.utils import SHMHandle as SHMHandle, deserialize_model as deserialize_model, process_tensors_in_object as process_tensors_in_object, reconstruct_from_shared_memory as reconstruct_from_shared_memory, serialize_model as serialize_model

__all__ = ['BaseClientTrainer', 'FilteredDataset', 'ProcessPoolClientTrainer', 'ThreadPoolClientTrainer', 'ModelSelector', 'PartitionedDataset', 'BaseServerHandler', 'serialize_model', 'deserialize_model', 'process_tensors_in_object', 'reconstruct_from_shared_memory', 'SHMHandle']



import numpy as np
import random
import torch
from dataclasses import dataclass

def setup_reproducibility(seed: int) -> None: ...

@dataclass
class RNGSuite:
    python: random.Random
    numpy: np.random.Generator
    torch_cpu: torch.Generator
    torch_cuda: torch.Generator | None = ...

def create_rng_suite(seed: int) -> RNGSuite: ...



import threading
from collections import defaultdict
from collections.abc import Iterable
from concurrent.futures import Future, as_completed
from copy import deepcopy
from dataclasses import dataclass
from pathlib import Path

import torch
import torch.nn.functional as F
from blazefl.core import (
    BaseServerHandler,
    FilteredDataset,
    ThreadPoolClientTrainer,
)
from blazefl.reproducibility import RNGSuite, create_rng_suite
from torch.utils.data import DataLoader, Subset
from tqdm import tqdm

from dataset import DSFLPartitionedDataset, DSFLPartitionType
from models import DSFLModelName, DSFLModelSelector


@dataclass
class DSFLUplinkPackage:
    cid: int
    soft_labels: torch.Tensor
    indices: torch.Tensor
    metadata: dict


@dataclass
class DSFLDownlinkPackage:
    soft_labels: torch.Tensor | None
    indices: torch.Tensor | None
    next_indices: torch.Tensor


class DSFLBaseServerHandler(BaseServerHandler[DSFLUplinkPackage, DSFLDownlinkPackage]):
    def __init__(
        self,
        model_selector: DSFLModelSelector,
        model_name: DSFLModelName,
        dataset: DSFLPartitionedDataset,
        global_round: int,
        num_clients: int,
        sample_ratio: float,
        device: str,
        kd_epochs: int,
        kd_batch_size: int,
        kd_lr: float,
        era_temperature: float,
        open_size_per_round: int,
        seed: int,
    ) -> None:
        self.model = model_selector.select_model(model_name)
        self.dataset = dataset
        self.global_round = global_round
        self.num_clients = num_clients
        self.sample_ratio = sample_ratio
        self.device = device
        self.kd_epochs = kd_epochs
        self.kd_batch_size = kd_batch_size
        self.kd_lr = kd_lr
        self.era_temperature = era_temperature
        self.open_size_per_round = open_size_per_round
        self.seed = seed

        self.kd_optimizer = torch.optim.SGD(self.model.parameters(), lr=kd_lr)
        self.client_buffer_cache: list[DSFLUplinkPackage] = []
        self.global_soft_labels: torch.Tensor | None = None
        self.global_indices: torch.Tensor | None = None
        self.num_clients_per_round = int(self.num_clients * self.sample_ratio)
        self.round = 0

        self.rng_suite = create_rng_suite(seed)

    def sample_clients(self) -> list[int]:
        sampled_clients = self.rng_suite.python.sample(
            range(self.num_clients), self.num_clients_per_round
        )

        return sorted(sampled_clients)

    def get_next_indices(self) -> torch.Tensor:
        shuffled_indices = torch.randperm(
            self.dataset.open_size, generator=self.rng_suite.torch_cpu
        )
        return shuffled_indices[: self.open_size_per_round]

    def if_stop(self) -> bool:
        return self.round >= self.global_round

    def load(self, payload: DSFLUplinkPackage) -> bool:
        self.client_buffer_cache.append(payload)

        if len(self.client_buffer_cache) == self.num_clients_per_round:
            self.global_update(self.client_buffer_cache)
            self.round += 1
            self.client_buffer_cache = []
            return True
        else:
            return False

    def global_update(self, buffer) -> None:
        buffer.sort(key=lambda x: x.cid)
        soft_labels_list = [ele.soft_labels for ele in buffer]
        indices_list = [ele.indices for ele in buffer]
        self.metadata_list = [ele.metadata for ele in buffer]

        soft_labels_stack: defaultdict[int, list[torch.Tensor]] = defaultdict(
            list[torch.Tensor]
        )
        for soft_labels, indices in zip(soft_labels_list, indices_list, strict=True):
            assert type(soft_labels) is torch.Tensor and type(indices) is torch.Tensor
            for soft_label, index in zip(soft_labels, indices, strict=True):
                soft_labels_stack[int(index.item())].append(soft_label)

        global_soft_labels: list[torch.Tensor] = []
        global_indices: list[int] = []
        for indices, soft_labels in sorted(
            soft_labels_stack.items(), key=lambda x: x[0]
        ):
            global_indices.append(indices)
            mean_soft_labels = torch.mean(torch.stack(soft_labels), dim=0)
            # Entropy Reduction Aggregation (ERA)
            era_soft_labels = F.softmax(mean_soft_labels / self.era_temperature, dim=0)
            global_soft_labels.append(era_soft_labels)

        open_dataset = self.dataset.get_dataset(type_=DSFLPartitionType.OPEN, cid=None)
        open_loader = DataLoader(
            Subset(open_dataset, global_indices),
            batch_size=self.kd_batch_size,
        )
        DSFLBaseServerHandler.distill(
            self.model,
            self.kd_optimizer,
            open_loader,
            global_soft_labels,
            self.kd_epochs,
            self.kd_batch_size,
            self.device,
            stop_event=None,
        )

        self.global_soft_labels = torch.stack(global_soft_labels)
        self.global_indices = torch.tensor(global_indices)

    @staticmethod
    def distill(
        model: torch.nn.Module,
        optimizer: torch.optim.Optimizer,
        open_loader: DataLoader,
        global_soft_labels: list[torch.Tensor],
        kd_epochs: int,
        kd_batch_size: int,
        device: str,
        stop_event: threading.Event | None,
    ) -> None:
        model.to(device)
        model.train()
        global_soft_label_loader = DataLoader(
            FilteredDataset(
                indices=list(range(len(global_soft_labels))),
                original_data=global_soft_labels,
            ),
            batch_size=kd_batch_size,
        )
        for _ in range(kd_epochs):
            if stop_event is not None and stop_event.is_set():
                break
            for data, soft_label in zip(
                open_loader, global_soft_label_loader, strict=True
            ):
                data = data.to(device)
                soft_label = soft_label.to(device).squeeze(1)

                output = model(data)
                loss = F.kl_div(
                    F.log_softmax(output, dim=1), soft_label, reduction="batchmean"
                )

                optimizer.zero_grad()
                loss.backward()
                optimizer.step()
        return

    @staticmethod
    def evaulate(
        model: torch.nn.Module, test_loader: DataLoader, device: str
    ) -> tuple[float, float]:
        model.to(device)
        model.eval()
        criterion = torch.nn.CrossEntropyLoss()

        total_loss = 0.0
        total_correct = 0
        total_samples = 0
        with torch.no_grad():
            for inputs, labels in test_loader:
                inputs = inputs.to(device)
                labels = labels.to(device)

                outputs = model(inputs)
                loss = criterion(outputs, labels)

                _, predicted = torch.max(outputs, 1)
                correct = torch.sum(predicted.eq(labels)).item()

                batch_size = labels.size(0)
                total_loss += loss.item() * batch_size
                total_correct += int(correct)
                total_samples += batch_size

        avg_loss = total_loss / total_samples
        avg_acc = total_correct / total_samples

        return avg_loss, avg_acc

    def get_summary(self) -> dict[str, float]:
        server_loss, server_acc = DSFLBaseServerHandler.evaulate(
            self.model,
            self.dataset.get_dataloader(
                type_=DSFLPartitionType.TEST,
                cid=None,
                batch_size=self.kd_batch_size,
            ),
            self.device,
        )
        client_loss = sum(m["loss"] for m in self.metadata_list) / len(
            self.metadata_list
        )
        client_acc = sum(m["acc"] for m in self.metadata_list) / len(self.metadata_list)
        return {
            "server_acc": server_acc,
            "server_loss": server_loss,
            "client_acc": client_acc,
            "client_loss": client_loss,
        }

    def downlink_package(self) -> DSFLDownlinkPackage:
        next_indices = self.get_next_indices()
        return DSFLDownlinkPackage(
            self.global_soft_labels, self.global_indices, next_indices
        )


@dataclass
class DSFLClientConfig:
    model_selector: DSFLModelSelector
    model_name: DSFLModelName
    dataset: DSFLPartitionedDataset
    epochs: int
    batch_size: int
    lr: float
    kd_epochs: int
    kd_batch_size: int
    kd_lr: float
    cid: int
    seed: int
    state_path: Path


@dataclass
class DSFLClientState:
    random: RNGSuite
    model: dict[str, torch.Tensor]
    optimizer: dict[str, torch.Tensor]
    kd_optimizer: dict[str, torch.Tensor] | None


class DSFLThreadPoolClientTrainer(
    ThreadPoolClientTrainer[DSFLUplinkPackage, DSFLDownlinkPackage]
):
    def __init__(
        self,
        model_selector: DSFLModelSelector,
        model_name: DSFLModelName,
        state_dir: Path,
        dataset: DSFLPartitionedDataset,
        device: str,
        num_clients: int,
        epochs: int,
        batch_size: int,
        lr: float,
        kd_epochs: int,
        kd_batch_size: int,
        kd_lr: float,
        seed: int,
        num_parallels: int,
    ) -> None:
        self.num_parallels = num_parallels
        self.device = device
        if self.device == "cuda":
            self.device_count = torch.cuda.device_count()
        self.cache: list[DSFLUplinkPackage] = []

        self.model_selector = model_selector
        self.model_name = model_name
        self.state_dir = state_dir
        self.state_dir.mkdir(parents=True, exist_ok=True)
        self.dataset = dataset
        self.epochs = epochs
        self.batch_size = batch_size
        self.lr = lr
        self.kd_epochs = kd_epochs
        self.kd_batch_size = kd_batch_size
        self.kd_lr = kd_lr
        self.device = device
        self.num_clients = num_clients
        self.seed = seed
        self.stop_event = threading.Event()

        self.client_states: dict[int, DSFLClientState] = {}
        self.lock = threading.Lock()

    def progress_fn(
        self, it: list[Future[DSFLUplinkPackage]]
    ) -> Iterable[Future[DSFLUplinkPackage]]:
        return tqdm(as_completed(it), total=len(it), desc="Client", leave=False)

    def worker(
        self,
        cid: int,
        device: str,
        payload: DSFLDownlinkPackage,
        stop_event: threading.Event,
    ) -> DSFLUplinkPackage:
        model = self.model_selector.select_model(self.model_name)
        optimizer = torch.optim.SGD(model.parameters(), lr=self.lr)
        kd_optimizer: torch.optim.SGD | None = None

        with self.lock:
            if cid in self.client_states:
                state = self.client_states[cid]
                rng_suite = state.random
                model.load_state_dict(state.model)
                optimizer.load_state_dict(state.optimizer)
                if state.kd_optimizer is not None:
                    kd_optimizer = torch.optim.SGD(model.parameters(), lr=self.kd_lr)
                    assert kd_optimizer is not None
                    kd_optimizer.load_state_dict(state.kd_optimizer)
            else:
                rng_suite = create_rng_suite(self.seed + cid)

        # Distill
        open_dataset = self.dataset.get_dataset(type_=DSFLPartitionType.OPEN, cid=None)
        if payload.indices is not None and payload.soft_labels is not None:
            global_soft_labels = list(torch.unbind(payload.soft_labels, dim=0))
            global_indices = payload.indices.tolist()
            if kd_optimizer is None:
                kd_optimizer = torch.optim.SGD(model.parameters(), lr=self.kd_lr)

            open_loader = DataLoader(
                Subset(open_dataset, global_indices),
                batch_size=self.kd_batch_size,
            )
            DSFLBaseServerHandler.distill(
                model=model,
                optimizer=kd_optimizer,
                open_loader=open_loader,
                global_soft_labels=global_soft_labels,
                kd_epochs=self.kd_epochs,
                kd_batch_size=self.kd_batch_size,
                device=device,
                stop_event=stop_event,
            )

        # Train
        train_loader = self.dataset.get_dataloader(
            type_=DSFLPartitionType.TRAIN,
            cid=cid,
            batch_size=self.batch_size,
            generator=rng_suite.torch_cpu,
        )
        DSFLThreadPoolClientTrainer.train(
            model=model,
            optimizer=optimizer,
            train_loader=train_loader,
            device=device,
            epochs=self.epochs,
            stop_event=stop_event,
        )
        self.dataset.set_dataset(
            dataset=train_loader.dataset, type_=DSFLPartitionType.TRAIN, cid=cid
        )

        # Predict
        open_loader = DataLoader(
            Subset(open_dataset, payload.next_indices.tolist()),
            batch_size=self.batch_size,
        )
        soft_labels = DSFLThreadPoolClientTrainer.predict(
            model=model,
            open_loader=open_loader,
            device=device,
        )

        # Evaluate
        test_loader = self.dataset.get_dataloader(
            type_=DSFLPartitionType.TEST,
            cid=cid,
            batch_size=self.batch_size,
        )
        loss, acc = DSFLBaseServerHandler.evaulate(
            model=model,
            test_loader=test_loader,
            device=device,
        )

        package = DSFLUplinkPackage(
            cid=cid,
            soft_labels=soft_labels,
            indices=payload.next_indices,
            metadata={"loss": loss, "acc": acc},
        )

        with self.lock:
            self.client_states[cid] = DSFLClientState(
                random=rng_suite,
                model=model.state_dict(),
                optimizer=optimizer.state_dict(),
                kd_optimizer=kd_optimizer.state_dict() if kd_optimizer else None,
            )

        return package

    @staticmethod
    def train(
        model: torch.nn.Module,
        optimizer: torch.optim.Optimizer,
        train_loader: DataLoader,
        device: str,
        epochs: int,
        stop_event: threading.Event,
    ) -> None:
        model.to(device)
        model.train()
        criterion = torch.nn.CrossEntropyLoss()

        for _ in range(epochs):
            if stop_event.is_set():
                break
            for data, target in train_loader:
                data = data.to(device)
                target = target.to(device)

                output = model(data)
                loss = criterion(output, target)

                optimizer.zero_grad()
                loss.backward()
                optimizer.step()
        return

    @staticmethod
    def predict(
        model: torch.nn.Module,
        open_loader: DataLoader,
        device: str,
    ) -> torch.Tensor:
        model.to(device)
        model.eval()

        soft_labels_list = []
        with torch.no_grad():
            for data in open_loader:
                data = data.to(device)

                output = model(data)
                soft_label = F.softmax(output, dim=1)

                soft_labels_list.append(soft_label.detach())

        soft_labels = torch.cat(soft_labels_list, dim=0)
        return soft_labels.cpu()

    def uplink_package(self) -> list[DSFLUplinkPackage]:
        package = deepcopy(self.cache)
        self.cache = []
        return package



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import importlib.metadata

project = "blazefl"
author = "kitsuyaazuma"
copyright = f"2025, {author}"
release = importlib.metadata.version(project)

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

import sys  # noqa: E402
from pathlib import Path  # noqa: E402

sys.path.insert(0, str(Path("../../src").resolve()))

extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.viewcode",
]

autosummary_generate = True
autosummary_ignore_module_all = False

templates_path = ["_templates"]
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_book_theme"
html_logo = "_static/logo.svg"
html_favicon = "_static/favicon.ico"
html_static_path = ["_static"]



import torch
from dataclasses import dataclass
from typing import Literal, TypeVar

@dataclass(frozen=True)
class SHMHandle: ...
T = TypeVar('T')

def process_tensors_in_object(obj: T, mode: Literal['move', 'replace'], max_depth: int = 10) -> T: ...
def reconstruct_from_shared_memory(handle_obj: T, shm_obj: T) -> T: ...
def serialize_model(model: torch.nn.Module, cpu: bool = True) -> torch.Tensor: ...
def deserialize_model(model: torch.nn.Module, serialized_parameters: torch.Tensor) -> None: ...



# ruff: noqa: E402
import logging
import warnings

from pydantic.warnings import UnsupportedFieldAttributeWarning

warnings.filterwarnings("ignore", category=UnsupportedFieldAttributeWarning)

import sys
from datetime import datetime
from pathlib import Path
from typing import Annotated

import torch
import torch.multiprocessing as mp
import typer
import wandb
from blazefl.contrib import (
    FedAvgBaseClientTrainer,
    FedAvgBaseServerHandler,
    FedAvgProcessPoolClientTrainer,
    FedAvgThreadPoolClientTrainer,
)
from blazefl.reproducibility import setup_reproducibility

from dataset import PartitionedCIFAR10
from models import FedAvgModelSelector
from models.selector import FedAvgModelName


class FedAvgPipeline:
    def __init__(
        self,
        handler: FedAvgBaseServerHandler,
        trainer: FedAvgBaseClientTrainer
        | FedAvgProcessPoolClientTrainer
        | FedAvgThreadPoolClientTrainer,
        run: wandb.Run,
    ) -> None:
        self.handler = handler
        self.trainer = trainer
        self.run = run

    def main(self):
        while not self.handler.if_stop():
            round_ = self.handler.round
            # server side
            sampled_clients = self.handler.sample_clients()
            broadcast = self.handler.downlink_package()

            # client side
            self.trainer.local_process(broadcast, sampled_clients)
            uploads = self.trainer.uplink_package()

            # server side
            for pack in uploads:
                self.handler.load(pack)

            summary = self.handler.get_summary()
            self.run.log(summary, step=round_)
            formatted_summary = ", ".join(f"{k}: {v:.3f}" for k, v in summary.items())
            logging.info(f"round: {round_}, {formatted_summary}")

        logging.info("done!")


def setup_logging() -> None:
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.INFO)
    formatter = logging.Formatter(
        "%(asctime)s [%(levelname)s] %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)


def main(
    model_name: Annotated[
        FedAvgModelName, typer.Option(help="Name of the model to be used.")
    ] = FedAvgModelName.CNN,
    num_clients: Annotated[
        int, typer.Option(help="Total number of clients in the federation.")
    ] = 100,
    global_round: Annotated[
        int, typer.Option(help="Total number of federated learning rounds.")
    ] = 5,
    sample_ratio: Annotated[
        float, typer.Option(help="Fraction of clients to sample in each round.")
    ] = 1.0,
    partition: Annotated[
        str,
        typer.Option(
            help="Dataset partition strategy ('shards' or 'client_inner_dirichlet')."
        ),
    ] = "shards",
    num_shards: Annotated[
        int, typer.Option(help="Number of shards for shard-based partitioning.")
    ] = 200,
    dir_alpha: Annotated[
        float,
        typer.Option(help="Alpha for Dirichlet distribution based partitioning."),
    ] = 1.0,
    seed: Annotated[int, typer.Option(help="Seed for reproducibility.")] = 42,
    epochs: Annotated[
        int, typer.Option(help="Number of local training epochs per client.")
    ] = 5,
    lr: Annotated[
        float, typer.Option(help="Learning rate for the client optimizer.")
    ] = 0.1,
    batch_size: Annotated[
        int, typer.Option(help="Batch size for local training.")
    ] = 50,
    num_parallels: Annotated[
        int,
        typer.Option(help="Number of parallel processes for training."),
    ] = 10,
    dataset_root_dir: Annotated[
        Path, typer.Option(help="Root directory for the dataset.")
    ] = Path("/tmp/quickstart-fedavg/dataset"),
    state_dir_base: Annotated[
        Path, typer.Option(help="Directory path for saving data between processes.")
    ] = Path("/tmp/quickstart-fedavg/state"),
    execution_mode: Annotated[
        str,
        typer.Option(
            help=(
                "Execution mode: 'single-threaded', 'multi-process', or "
                "'multi-threaded'."
            )
        ),
    ] = "multi-threaded",
) -> None:
    device = "cpu"
    if torch.cuda.is_available():
        device = "cuda"
    elif torch.backends.mps.is_available():
        device = "mps"

    config = locals()
    run = wandb.init(mode="offline", config=config)

    setup_logging()
    logging.info("\n" + "\n".join([f"  {k}: {v}" for k, v in config.items()]))

    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    dataset_split_dir = dataset_root_dir / timestamp
    state_dir = state_dir_base / timestamp

    setup_reproducibility(seed)

    dataset = PartitionedCIFAR10(
        root=dataset_root_dir,
        path=dataset_split_dir,
        num_clients=num_clients,
        seed=seed,
        partition=partition,
        num_shards=num_shards,
        dir_alpha=dir_alpha,
    )
    model_selector = FedAvgModelSelector(num_classes=10, seed=seed)

    handler = FedAvgBaseServerHandler(
        model_selector=model_selector,
        model_name=model_name,
        dataset=dataset,
        global_round=global_round,
        num_clients=num_clients,
        device=device,
        sample_ratio=sample_ratio,
        batch_size=batch_size,
        seed=seed,
    )
    trainer: (
        FedAvgBaseClientTrainer
        | FedAvgProcessPoolClientTrainer
        | FedAvgThreadPoolClientTrainer
        | None
    ) = None
    match execution_mode:
        case "single-threaded":
            trainer = FedAvgBaseClientTrainer(
                model_selector=model_selector,
                model_name=model_name,
                dataset=dataset,
                device=device,
                num_clients=num_clients,
                epochs=epochs,
                lr=lr,
                batch_size=batch_size,
                seed=seed,
            )
        case "multi-process":
            trainer = FedAvgProcessPoolClientTrainer(
                model_selector=model_selector,
                model_name=model_name,
                dataset=dataset,
                state_dir=state_dir,
                seed=seed,
                device=device,
                num_clients=num_clients,
                epochs=epochs,
                lr=lr,
                batch_size=batch_size,
                num_parallels=num_parallels,
            )
        case "multi-threaded":
            trainer = FedAvgThreadPoolClientTrainer(
                model_selector=model_selector,
                model_name=model_name,
                dataset=dataset,
                seed=seed,
                device=device,
                num_clients=num_clients,
                epochs=epochs,
                lr=lr,
                batch_size=batch_size,
                num_parallels=num_parallels,
            )
        case _:
            raise ValueError(f"Invalid execution mode: {execution_mode}")
    pipeline = FedAvgPipeline(handler=handler, trainer=trainer, run=run)
    try:
        pipeline.main()
    except KeyboardInterrupt:
        logging.info("KeyboardInterrupt")


if __name__ == "__main__":
    # NOTE: To use CUDA with multiprocessing, you must use the 'spawn' start method
    mp.set_start_method("spawn")

    typer.run(main)



import logging
import math
import os
import re
import statistics
import subprocess
from datetime import datetime
from pathlib import Path
from typing import Literal, NamedTuple

import toml
import torch
import typer
from rich.console import Console
from rich.logging import RichHandler
from rich.progress import track
from rich.table import Table

LOG_DIR = Path(__file__).parent / "logs"
LOG_DIR.mkdir(exist_ok=True)
LOG_FILE = LOG_DIR / f"{datetime.now():%Y%m%d_%H%M%S}.log"


def setup_logging() -> None:
    console_handler = RichHandler(markup=True, rich_tracebacks=True)
    console_handler.setFormatter(logging.Formatter("%(message)s", datefmt="[%X]"))
    console_handler.setLevel(logging.INFO)

    file_handler = logging.FileHandler(LOG_FILE)
    file_handler.setFormatter(
        logging.Formatter("%(asctime)s - %(levelname)s - %(message)s", datefmt="[%X]")
    )

    logging.basicConfig(
        level=logging.DEBUG,
        handlers=[console_handler, file_handler],
    )


def run_benchmark(command: str, num_runs: int) -> list[float]:
    execution_times: list[float] = []
    for i in track(range(num_runs), description="Benchmark Progress"):
        logging.info(f"Running benchmark (run {i + 1}/{num_runs}): {command}")
        try:
            result = subprocess.run(
                command, shell=True, capture_output=True, text=True, check=True
            )

            if result.stdout:
                logging.debug(f"Run {i + 1} stdout:\n{result.stdout.strip()}")
            if result.stderr:
                logging.debug(f"Run {i + 1} stderr:\n{result.stderr.strip()}")

            output = result.stdout + result.stderr
            match = re.search(r"BENCHMARK_RESULT_TIME: ([\d.]+)", output)
            if match:
                time_taken = float(match.group(1))
                execution_times.append(time_taken)
                logging.info(
                    f"Run {i + 1}/{num_runs} finished in {time_taken:.4f} seconds."
                )
            else:
                logging.warning(
                    f"Could not find benchmark result in output for run {i + 1}."
                )
        except subprocess.CalledProcessError as e:
            logging.error(f"Command failed for run {i + 1}/{num_runs}")
            logging.error(f"Return code: {e.returncode}")

            if e.stdout:
                logging.error(f"stdout:\n{e.stdout.strip()}")
            if e.stderr:
                logging.error(f"stderr:\n{e.stderr.strip()}")
            continue

    return execution_times


class Result(NamedTuple):
    method: str
    avg_time: float
    std_time: float


def display_results(title: str, results: list[Result]) -> None:
    console = Console()

    table = Table(title=title, show_header=True, header_style="bold magenta")
    table.add_column("Method", style="dim")
    table.add_column("Execution Time (s)", justify="right")
    for result in results:
        row = f"{result.avg_time:.4f} ± {result.std_time:.4f}"
        table.add_row(result.method, row)
        logging.info(f"Result - {result.method}: {row}")

    console.print(table)


def main(num_runs: int = 3, model_name: Literal["CNN", "RESNET18"] = "CNN") -> None:
    logging.info("Starting benchmark...")
    cpu_count = os.cpu_count() or 1
    gpu_count = torch.cuda.device_count() if torch.cuda.is_available() else 0

    num_parallels: list[int] = [2**i for i in range(int(math.log2(cpu_count) + 1))]

    for num_parallel in num_parallels:
        results: list[Result] = []

        execution_modes = ["MULTI_THREADED", "MULTI_PROCESS"]
        for mode in execution_modes:
            blazefl_command = (
                "cd blazefl-case && "
                "uv run python "
                f"{'-Xgil=0' if mode == 'MULTI_THREADED' else ''} main.py "
                f"--model-name={model_name} "
                f"--execution-mode={mode} "
                f"--num-parallels={num_parallel} "
            )
            blazefl_times = run_benchmark(blazefl_command, num_runs)
            if blazefl_times:
                result = Result(
                    method="BlazeFL",
                    avg_time=statistics.mean(blazefl_times),
                    std_time=statistics.stdev(blazefl_times)
                    if len(blazefl_times) > 1
                    else 0.0,
                )
                results.append(result)

        client_cpus = cpu_count // num_parallel
        client_gpus = gpu_count / num_parallel

        # Update Flower config
        config_tmpl_path = Path("flower-case/config.toml.tmpl")
        config_path = Path("flower-case/config.toml")
        if config_tmpl_path.exists():
            config = toml.load(config_tmpl_path)
            config["superlink"]["local"]["options"]["backend"]["client-resources"][
                "num-cpus"
            ] = client_cpus
            config["superlink"]["local"]["options"]["backend"]["client-resources"][
                "num-gpus"
            ] = float(client_gpus)
            with open(config_path, "w") as f:
                toml.dump(config, f)

        flower_command = (
            "cd flower-case && "
            "FLWR_HOME=$(pwd) "
            "uv run flwr run . local "
            f"--run-config 'model-name=\"{model_name}\"' "
            "&& cd .."
        )
        flower_times = run_benchmark(flower_command, num_runs)
        if flower_times:
            result = Result(
                method="Flower",
                avg_time=statistics.mean(flower_times),
                std_time=statistics.stdev(flower_times)
                if len(flower_times) > 1
                else 0.0,
            )
            results.append(result)

        display_results(f"FedAvg Benchmark ({num_parallel=})", results)
    logging.info("Benchmark finished.")


if __name__ == "__main__":
    setup_logging()
    typer.run(main)



from dataclasses import dataclass, field
from typing import Any

import pytest
import torch

from src.blazefl.core import (
    SHMHandle,
    deserialize_model,
    process_tensors_in_object,
    reconstruct_from_shared_memory,
    serialize_model,
)


@pytest.fixture
def simple_model() -> torch.nn.Module:
    model = torch.nn.Sequential(
        torch.nn.Linear(4, 3), torch.nn.ReLU(), torch.nn.Linear(3, 1)
    )
    return model


def test_serialize_deserialize_cpu(simple_model: torch.nn.Module) -> None:
    original_params = [p.clone() for p in simple_model.parameters()]
    serialized = serialize_model(simple_model, cpu=True)

    assert serialized.device.type == "cpu"

    total_numel = sum(p.numel() for p in simple_model.parameters())
    assert serialized.numel() == total_numel

    for p in simple_model.parameters():
        p.data.normal_()

    deserialize_model(simple_model, serialized)

    for orig_p, new_p in zip(original_params, simple_model.parameters(), strict=True):
        assert torch.allclose(orig_p, new_p), "Parameters did not restore correctly."


def test_serialize_deserialize_gpu(simple_model: torch.nn.Module) -> None:
    if torch.cuda.is_available():
        simple_model = simple_model.to("cuda")
        original_params = [p.clone() for p in simple_model.parameters()]

        serialized_gpu = serialize_model(simple_model, cpu=False)
        assert serialized_gpu.device.type == "cuda"

        serialized_cpu = serialize_model(simple_model, cpu=True)
        assert serialized_cpu.device.type == "cpu"

        for p in simple_model.parameters():
            p.data.normal_()

        simple_model = simple_model.cpu()
        deserialize_model(simple_model, serialized_cpu)

        for orig_p, new_p in zip(
            original_params, simple_model.parameters(), strict=True
        ):
            assert torch.allclose(orig_p.cpu(), new_p), (
                "Parameters did not restore correctly."
            )


@dataclass
class NestedData:
    tensor: torch.Tensor
    id: int


@dataclass
class ComplexData:
    tensor_a: torch.Tensor
    tensor_b: torch.Tensor
    metadata: dict[str, Any]
    nested: NestedData
    # Field for creating circular references
    parent: Any = field(default=None, repr=False)


@pytest.fixture
def complex_object() -> ComplexData:
    """Creates a nested dataclass object for testing."""
    obj = ComplexData(
        tensor_a=torch.randn(2, 2),
        tensor_b=torch.ones(5),
        metadata={"name": "test_object", "value": 123},
        nested=NestedData(tensor=torch.arange(4), id=1),
    )
    return obj


@pytest.fixture
def circular_object() -> ComplexData:
    """Creates an object with a circular reference."""
    obj = ComplexData(
        tensor_a=torch.randn(2, 2),
        tensor_b=torch.ones(5),
        metadata={"name": "circular_test"},
        nested=NestedData(tensor=torch.arange(4), id=2),
    )
    # Create the circular reference
    obj.parent = obj
    return obj


def test_process_tensors_move_mode(complex_object: ComplexData):
    """
    Tests that 'move' mode modifies tensors in-place and doesn't change the object ID.
    """
    original_id = id(complex_object)
    # Mock share_memory_ to check if it's called
    call_count = 0
    original_share_memory = torch.Tensor.share_memory_

    def mocked_share_memory(self):
        nonlocal call_count
        call_count += 1

    torch.Tensor.share_memory_ = mocked_share_memory

    result = process_tensors_in_object(complex_object, mode="move", max_depth=10)

    # Restore original method
    torch.Tensor.share_memory_ = original_share_memory

    assert id(result) == original_id, "Object ID should be the same in 'move' mode"
    assert call_count == 3, "share_memory_() should be called on all 3 tensors"


def test_process_tensors_replace_mode(complex_object: ComplexData):
    """
    Tests that 'replace' mode creates a new object and replaces tensors with SHMHandle.
    """
    original_id = id(complex_object)
    handle_package = process_tensors_in_object(
        complex_object, mode="replace", max_depth=10
    )

    assert id(handle_package) != original_id, (
        "Object ID should be different in 'replace' mode"
    )
    assert isinstance(handle_package.tensor_a, SHMHandle)
    assert isinstance(handle_package.tensor_b, SHMHandle)
    assert isinstance(handle_package.nested.tensor, SHMHandle)
    assert handle_package.metadata["name"] == "test_object", (
        "Non-tensor data should be preserved"
    )


def test_reconstruct_from_shared_memory(complex_object: ComplexData):
    """
    Tests the reconstruction logic, ensuring non-tensor data from the handle object
    and tensor data from the shm_buffer are correctly combined.
    """
    # 1. Create the shared memory buffer (pretending it's in SHM)
    shm_buffer = complex_object

    # 2. Create the handle package
    handle_package = process_tensors_in_object(shm_buffer, mode="replace", max_depth=10)

    # 3. Modify a non-tensor value in the handle package (simulating a worker update)
    handle_package.metadata["value"] = 999
    handle_package.nested.id = 5

    # 4. Reconstruct the final object
    final_package = reconstruct_from_shared_memory(handle_package, shm_buffer)

    # 5. Assertions
    assert torch.allclose(final_package.tensor_a, shm_buffer.tensor_a), (
        "Tensor A should be from shm_buffer"
    )
    assert torch.allclose(final_package.nested.tensor, shm_buffer.nested.tensor), (
        "Nested tensor should be from shm_buffer"
    )
    assert final_package.metadata["value"] == 999, (
        "Metadata update from handle_package should be preserved"
    )
    assert final_package.nested.id == 5, (
        "Nested ID update from handle_package should be preserved"
    )
    assert not isinstance(final_package.tensor_a, SHMHandle), (
        "Final package should not contain handles"
    )


def test_circular_reference_handling(circular_object: ComplexData):
    """
    Tests that both functions complete without infinite recursion when a circular
    reference is present.
    """
    # Test 'move' mode
    try:
        process_tensors_in_object(circular_object, mode="move", max_depth=10)
    except RecursionError:
        pytest.fail(
            "process_tensors_in_object in 'move' mode failed on circular reference"
        )

    # Test 'replace' mode
    try:
        handle_package = process_tensors_in_object(
            circular_object, mode="replace", max_depth=10
        )
    except RecursionError:
        pytest.fail(
            "process_tensors_in_object in 'replace' mode failed on circular reference"
        )

    # Test reconstruction
    try:
        reconstruct_from_shared_memory(handle_package, circular_object)
    except RecursionError:
        pytest.fail("reconstruct_from_shared_memory failed on circular reference")


def test_max_depth_parameter(complex_object: ComplexData):
    """
    Tests that recursion stops at the specified max_depth.
    """
    # With max_depth=1, only top-level tensors should be replaced.
    handle_package = process_tensors_in_object(
        complex_object, mode="replace", max_depth=2
    )  # depth starts at 0, __dict__ is 1

    assert isinstance(handle_package.tensor_a, SHMHandle)
    # The nested tensor should NOT be a handle, as it's beyond max_depth
    assert isinstance(handle_package.nested.tensor, torch.Tensor)



from typing import Protocol, TypeVar

UplinkPackage = TypeVar("UplinkPackage")
DownlinkPackage = TypeVar("DownlinkPackage", covariant=True)


class BaseServerHandler(Protocol[UplinkPackage, DownlinkPackage]):
    """
    Abstract base class for server-side operations in federated learning.

    This class defines the essential methods that a server handler must implement
    to manage communication and coordination with clients during federated learning
    processes. It uses generic types `UplinkPackage` and `DownlinkPackage` to
    define the types of data exchanged between the server and clients.

    Raises:
        NotImplementedError: If any of the methods are not implemented in a subclass.
    """

    def downlink_package(self) -> DownlinkPackage:
        """
        Prepare the data package to be sent from the server to clients.

        Returns:
            DownlinkPackage: The data package intended for client consumption.
        """
        ...

    def sample_clients(self) -> list[int]:
        """
        Select a list of client IDs to participate in the current training round.

        Returns:
            list[int]: A list of selected client IDs.
        """
        ...

    def if_stop(self) -> bool:
        """
        Determine whether the federated learning process should be terminated.

        Returns:
            bool: True if the process should stop; False otherwise.
        """
        ...

    def global_update(self, buffer: list[UplinkPackage]) -> None:
        """
        Update the global model based on the aggregated data from clients.

        Args:
            buffer (list[UplinkPackage]): A list containing data sent by clients,
            typically representing model updates or gradients.

        Returns:
            None
        """
        ...

    def load(self, payload: UplinkPackage) -> bool:
        """
        Load a given payload into the server's state.

        Args:
            payload (UplinkPackage): The data to be loaded into the server.

        Returns:
            bool: True if the payload was successfully loaded; False otherwise.
        """
        ...



import torch
from datasets import load_dataset
from flwr_datasets import FederatedDataset
from flwr_datasets.partitioner import PathologicalPartitioner
from torch.utils.data import DataLoader
from torchvision.transforms import Compose, Normalize, ToTensor

fds = None  # Cache FederatedDataset

pytorch_transforms = Compose([ToTensor(), Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))])


def apply_transforms(batch):
    """Apply transforms to the partition from FederatedDataset."""
    batch["img"] = [pytorch_transforms(img) for img in batch["img"]]
    return batch


def load_data(partition_id: int, num_partitions: int, batch_size: int):
    """Load partition CIFAR10 data."""
    # Only initialize `FederatedDataset` once
    global fds
    if fds is None:
        partitioner = PathologicalPartitioner(
            num_partitions=num_partitions,
            partition_by="label",
            num_classes_per_partition=2,
        )
        fds = FederatedDataset(
            dataset="uoft-cs/cifar10",
            partitioners={"train": partitioner},
        )
    partition = fds.load_partition(partition_id).with_transform(apply_transforms)
    trainloader = DataLoader(
        partition,  # pyright: ignore[reportArgumentType]
        batch_size=batch_size,
        shuffle=True,
    )
    return trainloader


def load_centralized_dataset(batch_size: int):
    """Load test set and return dataloader."""
    # Load entire test set
    test_dataset = load_dataset("uoft-cs/cifar10", split="test")
    dataset = test_dataset.with_format("torch").with_transform(apply_transforms)  # pyright: ignore[reportAttributeAccessIssue]
    return DataLoader(dataset, batch_size=batch_size)  # pyright: ignore[reportArgumentType]


def train(net, trainloader, epochs, lr, device):
    """Train the model on the training set."""
    net.to(device)  # move model to GPU if available
    criterion = torch.nn.CrossEntropyLoss().to(device)
    optimizer = torch.optim.SGD(net.parameters(), lr=lr)
    net.train()
    running_loss = 0.0
    for _ in range(epochs):
        for batch in trainloader:
            images = batch["img"].to(device)
            labels = batch["label"].to(device)
            optimizer.zero_grad()
            loss = criterion(net(images), labels)
            loss.backward()
            optimizer.step()
            running_loss += loss.item()
    avg_trainloss = running_loss / len(trainloader)
    return avg_trainloss


def test(net, testloader, device):
    """Validate the model on the test set."""
    net.to(device)
    criterion = torch.nn.CrossEntropyLoss()
    correct, loss = 0, 0.0
    with torch.no_grad():
        for batch in testloader:
            images = batch["img"].to(device)
            labels = batch["label"].to(device)
            outputs = net(images)
            loss += criterion(outputs, labels).item()
            correct += (torch.max(outputs.data, 1)[1] == labels).sum().item()
    accuracy = correct / len(testloader.dataset)
    loss = loss / len(testloader)
    return loss, accuracy



import random

import numpy as np
import torch

from src.blazefl.reproducibility import create_rng_suite


def test_rng_suite_creation_and_reproducibility() -> None:
    seed = 42

    rng_suite1 = create_rng_suite(seed)
    py_val1 = rng_suite1.python.random()
    np_val1 = rng_suite1.numpy.random()
    torch_cpu_val1 = torch.rand(1, generator=rng_suite1.torch_cpu)
    torch_cuda_val1 = None
    if rng_suite1.torch_cuda:
        torch_cuda_val1 = torch.rand(1, device="cuda", generator=rng_suite1.torch_cuda)

    rng_suite2 = create_rng_suite(seed)
    py_val2 = rng_suite2.python.random()
    np_val2 = rng_suite2.numpy.random()
    torch_cpu_val2 = torch.rand(1, generator=rng_suite2.torch_cpu)
    torch_cuda_val2 = None
    if rng_suite2.torch_cuda:
        torch_cuda_val2 = torch.rand(1, device="cuda", generator=rng_suite2.torch_cuda)

    assert py_val1 == py_val2
    assert np_val1 == np_val2
    assert torch.equal(torch_cpu_val1, torch_cpu_val2)
    if torch_cuda_val1 is not None:
        assert torch_cuda_val2 is not None
        assert torch.equal(torch_cuda_val1, torch_cuda_val2)

    rng_suite3 = create_rng_suite(seed + 1)
    py_val3 = rng_suite3.python.random()
    assert py_val1 != py_val3


def test_rng_suite_independence_from_global_state() -> None:
    seed = 99
    rng_suite = create_rng_suite(seed)

    py_ref = rng_suite.python.random()
    np_ref = rng_suite.numpy.random()
    torch_ref = torch.rand(1, generator=rng_suite.torch_cpu)

    random.seed(1)
    np.random.seed(1)
    torch.manual_seed(1)
    _ = random.random()
    _ = np.random.rand()
    _ = torch.rand(1)

    py_new = rng_suite.python.random()
    np_new = rng_suite.numpy.random()
    torch_new = torch.rand(1, generator=rng_suite.torch_cpu)

    assert py_ref != py_new
    assert np_ref != np_new
    assert not torch.equal(torch_ref, torch_new)

    rng_suite_recreated = create_rng_suite(seed)
    py_recreated = rng_suite_recreated.python.random()
    np_recreated = rng_suite_recreated.numpy.random()
    torch_recreated = torch.rand(1, generator=rng_suite_recreated.torch_cpu)

    assert py_ref == py_recreated
    assert np_ref == np_recreated
    assert torch.equal(torch_ref, torch_recreated)



import logging
import sys
import time
from datetime import datetime
from enum import StrEnum
from pathlib import Path

import torch
import torch.multiprocessing as mp
import typer
from blazefl.contrib import (
    FedAvgBaseClientTrainer,
    FedAvgBaseServerHandler,
    FedAvgProcessPoolClientTrainer,
    FedAvgThreadPoolClientTrainer,
)
from blazefl.reproducibility import setup_reproducibility
from dataset import PartitionedCIFAR10
from models import FedAvgModelName, FedAvgModelSelector


class FedAvgBenchmarkPipeline:
    def __init__(
        self,
        handler: FedAvgBaseServerHandler,
        trainer: FedAvgBaseClientTrainer
        | FedAvgProcessPoolClientTrainer
        | FedAvgThreadPoolClientTrainer,
    ) -> None:
        self.handler = handler
        self.trainer = trainer

    def main(self):
        start_time = time.perf_counter()
        while not self.handler.if_stop():
            round_ = self.handler.round
            # server side
            sampled_clients = self.handler.sample_clients()
            broadcast = self.handler.downlink_package()

            # client side
            self.trainer.local_process(broadcast, sampled_clients)
            uploads = self.trainer.uplink_package()

            # server side
            for pack in uploads:
                self.handler.load(pack)

            summary = self.handler.get_summary()
            formatted_summary = ", ".join(f"{k}: {v:.3f}" for k, v in summary.items())
            logging.info(f"round: {round_}, {formatted_summary}")

        end_time = time.perf_counter()
        logging.info(f"BENCHMARK_RESULT_TIME: {end_time - start_time:.4f}")


class ExecutionMode(StrEnum):
    SINGLE_THREADED = "SINGLE_THREADED"
    MULTI_PROCESS = "MULTI_PROCESS"
    MULTI_THREADED = "MULTI_THREADED"


EXECUTION_MODE = ExecutionMode.MULTI_THREADED


def main(
    model_name: FedAvgModelName = FedAvgModelName.CNN,
    num_clients: int = 100,
    global_round: int = 5,
    sample_ratio: float = 1.0,
    partition: str = "shards",
    num_shards: int = 200,
    dir_alpha: float = 1.0,
    seed: int = 42,
    epochs: int = 5,
    lr: float = 0.1,
    batch_size: int = 50,
    num_parallels: int = 10,
    dataset_root_dir: Path = Path("/tmp/blazefl-case/dataset"),
    state_dir_base: Path = Path("/tmp/blazefl-case/state"),
    execution_mode: ExecutionMode = EXECUTION_MODE,
):
    logging.basicConfig(level=logging.INFO, format="%(message)s")

    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    dataset_split_dir = dataset_root_dir / timestamp
    state_dir = state_dir_base / timestamp

    device = "cpu"
    if torch.cuda.is_available():
        device = "cuda"
    elif torch.backends.mps.is_available():
        device = "mps"

    setup_reproducibility(seed)

    dataset = PartitionedCIFAR10(
        root=dataset_root_dir,
        path=dataset_split_dir,
        num_clients=num_clients,
        seed=seed,
        partition=partition,
        num_shards=num_shards,
        dir_alpha=dir_alpha,
    )
    model_selector = FedAvgModelSelector(num_classes=10, seed=seed)

    handler = FedAvgBaseServerHandler(
        model_selector=model_selector,
        model_name=model_name,
        dataset=dataset,
        global_round=global_round,
        num_clients=num_clients,
        device=device,
        sample_ratio=sample_ratio,
        batch_size=batch_size,
        seed=seed,
    )
    trainer: (
        FedAvgBaseClientTrainer
        | FedAvgProcessPoolClientTrainer
        | FedAvgThreadPoolClientTrainer
        | None
    ) = None
    match execution_mode:
        case ExecutionMode.SINGLE_THREADED:
            trainer = FedAvgBaseClientTrainer(
                model_selector=model_selector,
                model_name=model_name,
                dataset=dataset,
                device=device,
                num_clients=num_clients,
                epochs=epochs,
                lr=lr,
                batch_size=batch_size,
                seed=seed,
            )
        case ExecutionMode.MULTI_PROCESS:
            trainer = FedAvgProcessPoolClientTrainer(
                model_selector=model_selector,
                model_name=model_name,
                dataset=dataset,
                state_dir=state_dir,
                seed=seed,
                device=device,
                num_clients=num_clients,
                epochs=epochs,
                lr=lr,
                batch_size=batch_size,
                num_parallels=num_parallels,
            )
        case ExecutionMode.MULTI_THREADED:
            assert not sys._is_gil_enabled()
            trainer = FedAvgThreadPoolClientTrainer(
                model_selector=model_selector,
                model_name=model_name,
                dataset=dataset,
                seed=seed,
                device=device,
                num_clients=num_clients,
                epochs=epochs,
                lr=lr,
                batch_size=batch_size,
                num_parallels=num_parallels,
            )
    pipeline = FedAvgBenchmarkPipeline(handler=handler, trainer=trainer)
    try:
        pipeline.main()
    except KeyboardInterrupt:
        logging.info("KeyboardInterrupt")


if __name__ == "__main__":
    # NOTE: To use CUDA with multiprocessing, you must use the 'spawn' start method
    mp.set_start_method("spawn")

    typer.run(main)



from collections.abc import Sized
from enum import StrEnum
from pathlib import Path

import torch
import torchvision
from blazefl.core import FilteredDataset, PartitionedDataset
from blazefl.reproducibility import create_rng_suite
from torch.utils.data import DataLoader, Dataset
from torchvision import transforms

from .functional import balance_split, client_inner_dirichlet_partition_faster


class DSFLPartitionType(StrEnum):
    TRAIN = "TRAIN"
    OPEN = "OPEN"
    TEST = "TEST"


class DSFLPartitionedDataset(PartitionedDataset[DSFLPartitionType]):
    def __init__(
        self,
        root: Path,
        path: Path,
        num_clients: int,
        seed: int,
        partition: str,
        open_size: int,
        dir_alpha: float,
    ) -> None:
        self.root = root
        self.path = path
        self.num_clients = num_clients
        self.seed = seed
        self.partition = partition
        self.dir_alpha = dir_alpha
        self.open_size = open_size

        self.rng_suite = create_rng_suite(seed)

        self.train_transform = transforms.Compose(
            [
                transforms.ToTensor(),
                transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5)),
            ]
        )
        self.test_transform = transforms.Compose(
            [
                transforms.ToTensor(),
                transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5)),
            ]
        )

        self._preprocess()

    def _preprocess(self):
        self.root.mkdir(parents=True, exist_ok=True)
        train_dataset = torchvision.datasets.CIFAR10(
            root=self.root,
            train=True,
            download=True,
        )
        open_dataset = torchvision.datasets.CIFAR100(
            root=self.root,
            train=True,
            download=True,
        )
        test_dataset = torchvision.datasets.CIFAR10(
            root=self.root,
            train=False,
            download=True,
        )
        for type_ in [ds.value for ds in DSFLPartitionType]:
            self.path.joinpath(type_).mkdir(parents=True)

        match self.partition:
            case "client_inner_dirichlet":
                client_dict, class_priors = client_inner_dirichlet_partition_faster(
                    targets=train_dataset.targets,
                    num_clients=self.num_clients,
                    num_classes=10,
                    dir_alpha=self.dir_alpha,
                    client_sample_nums=balance_split(
                        num_clients=self.num_clients,
                        num_samples=len(train_dataset.targets),
                    ),
                    verbose=False,
                    numpy_rng=self.rng_suite.numpy,
                )
                test_client_dict, _ = client_inner_dirichlet_partition_faster(
                    targets=test_dataset.targets,
                    num_clients=self.num_clients,
                    num_classes=10,
                    dir_alpha=self.dir_alpha,
                    client_sample_nums=balance_split(
                        num_clients=self.num_clients,
                        num_samples=len(test_dataset.targets),
                    ),
                    class_priors=class_priors,
                    verbose=False,
                    numpy_rng=self.rng_suite.numpy,
                )
            case _:
                raise ValueError(f"Invalid partition: {self.partition}")

        for cid, indices in client_dict.items():
            client_train_dataset = FilteredDataset(
                indices.tolist(),
                train_dataset.data,
                train_dataset.targets,
                transform=self.train_transform,
            )
            torch.save(client_train_dataset, self.path.joinpath("train", f"{cid}.pkl"))

        for cid, indices in test_client_dict.items():
            client_test_dataset = FilteredDataset(
                indices.tolist(),
                test_dataset.data,
                test_dataset.targets,
                transform=self.test_transform,
            )
            torch.save(client_test_dataset, self.path.joinpath("test", f"{cid}.pkl"))

        open_indices = self.rng_suite.numpy.choice(
            len(open_dataset),
            size=self.open_size,
        )
        torch.save(
            FilteredDataset(
                open_indices.tolist(),
                open_dataset.data,
                original_targets=None,
                transform=self.train_transform,
            ),
            self.path.joinpath("open.pkl"),
        )

        torch.save(
            FilteredDataset(
                list(range(len(test_dataset))),
                test_dataset.data,
                test_dataset.targets,
                transform=self.test_transform,
            ),
            self.path.joinpath("test", "default.pkl"),
        )

    def get_dataset(self, type_: DSFLPartitionType, cid: int | None) -> Dataset:
        match type_:
            case DSFLPartitionType.TRAIN:
                dataset = torch.load(
                    self.path.joinpath(type_, f"{cid}.pkl"),
                    weights_only=False,
                )
            case DSFLPartitionType.OPEN:
                dataset = torch.load(
                    self.path.joinpath(f"{type_}.pkl"),
                    weights_only=False,
                )
            case DSFLPartitionType.TEST:
                if cid is not None:
                    dataset = torch.load(
                        self.path.joinpath(type_, f"{cid}.pkl"),
                        weights_only=False,
                    )
                else:
                    dataset = torch.load(
                        self.path.joinpath(type_, "default.pkl"), weights_only=False
                    )
        assert isinstance(dataset, Dataset)
        return dataset

    def set_dataset(
        self, type_: DSFLPartitionType, cid: int | None, dataset: Dataset
    ) -> None:
        match type_:
            case DSFLPartitionType.TRAIN:
                torch.save(dataset, self.path.joinpath(type_, f"{cid}.pkl"))
            case DSFLPartitionType.OPEN:
                torch.save(dataset, self.path.joinpath(f"{type_}.pkl"))
            case DSFLPartitionType.TEST:
                if cid is not None:
                    torch.save(dataset, self.path.joinpath(type_, f"{cid}.pkl"))
                else:
                    torch.save(dataset, self.path.joinpath(type_, "default.pkl"))

    def get_dataloader(
        self,
        type_: DSFLPartitionType,
        cid: int | None,
        batch_size: int | None = None,
        generator: torch.Generator | None = None,
    ) -> DataLoader:
        dataset = self.get_dataset(type_, cid)
        assert isinstance(dataset, Sized)
        batch_size = len(dataset) if batch_size is None else batch_size
        data_loader = DataLoader(
            dataset,
            batch_size=batch_size,
            shuffle=type_ == DSFLPartitionType.TRAIN,
            generator=generator,
        )
        return data_loader
